from __future__ import annotations

from typing import Literal

from splink import ColumnExpression
from splink.internals.comparison_level_creator import ComparisonLevelCreator
from splink.internals.comparison_level_library import (
    AbsoluteTimeDifferenceLevel as SplinkAbsoluteTimeDifferenceLevel,
)
from splink.internals.comparison_level_library import (
    DateMetricType,
)
from splink.internals.comparison_level_library import (
    PairwiseStringDistanceFunctionLevel as SplinkPairwiseStringDistanceFunctionLevel,
)

from .column_expression import ColumnExpression as CHColumnExpression
from .dialect import ClickhouseDialect, SplinkDialect


class DistanceInKMLevel(ComparisonLevelCreator):
    calculation_method_lookup = {
        "great_circle": "greatCircleDistance",
        "wgs84": "geoDistance",
    }

    def __init__(
        self,
        lat_col: str | ColumnExpression,
        long_col: str | ColumnExpression,
        km_threshold: float,
        not_null: bool = False,
        calculation_method: Literal["great_circle", "wgs84"] = "wgs84",
    ):
        """
        Compare the distance between two latitude and longitude coördinate points
        as measured in kilometres.

        Calculation uses in-built Clickhouse functions, which support either
        using great circle distance, or the slightly more accurate WGS 84 method.

        Arguments:
            lat_col (str): The name of a latitude column or the respective array
                or struct column column containing the information
                For example: long_lat['lat'] or long_lat[0]
            long_col (str): The name of a longitudinal column or the respective array
                or struct column column containing the information, plus an index.
                For example: long_lat['long'] or long_lat[1]
            km_threshold (int): The total distance in kilometers to evaluate your
                comparisons against
            not_null (bool): If true, ensure no attempt is made to compute this if
                any inputs are null. This is only necessary if you are not
                capturing nulls elsewhere in your comparison level.
            calculation_method (str): The method to use to compute distances from
                latitude and longitude.
                Can be 'great_circle' for the great circle distance, based on a
                spherical model (Clickhouse `greatCircleDistance`), or 'wgs84'
                to use the more accurate WGS 84 Ellipsoid model
                (Clickhouse `geoDistance`). Default 'wgs84'.

        """
        if calculation_method not in self.calculation_method_lookup:
            valid_opts_str = "', '".join(self.calculation_method_lookup.keys())
            raise ValueError(
                f"Calculation method must be one of {valid_opts_str}, "
                f"not '{calculation_method}'"
            )
        self.lat_col_expression = ColumnExpression.instantiate_if_str(lat_col)
        self.long_col_expression = ColumnExpression.instantiate_if_str(long_col)

        self.km_threshold = km_threshold
        self.not_null = not_null
        self.calculation_method = calculation_method

    def create_sql(self, sql_dialect: SplinkDialect) -> str:
        if sql_dialect is not ClickhouseDialect():
            raise ValueError(
                "This version of `DistanceInKMLevel` is designed only for use "
                "with the Clickhouse dialect of SQL.\n"
                "Perhaps you meant to use "
                "`splink.comparison_level_library.DistanceInKMLevel`?"
            )
        self.lat_col_expression.sql_dialect = sql_dialect
        lat_col = self.lat_col_expression

        self.long_col_expression.sql_dialect = sql_dialect
        long_col = self.long_col_expression

        lat_l, lat_r = lat_col.name_l, lat_col.name_r
        long_l, long_r = long_col.name_l, long_col.name_r

        m_distance_function = self.calculation_method_lookup[self.calculation_method]

        # Clickhouse deals in metres
        m_threshold = self.km_threshold * 1000
        distance_km_sql = f"""
            {m_distance_function}({long_l}, {lat_l}, {long_r}, {lat_r}) <= {m_threshold}
        """.strip()

        if self.not_null:
            null_sql = " AND ".join(
                [f"{c} is not null" for c in [lat_r, lat_l, long_l, long_r]]
            )
            distance_km_sql = f"({null_sql}) AND {distance_km_sql}"

        return distance_km_sql

    def create_label_for_charts(self) -> str:
        return f"Distance less than {self.km_threshold}km"


class AbsoluteDateDifferenceLevel(SplinkAbsoluteTimeDifferenceLevel):
    def __init__(
        self,
        col_name: str | ColumnExpression,
        *,
        input_is_string: bool,
        threshold: float,
        metric: DateMetricType,
    ):
        """
        Computes the absolute time difference between two dates (total duration).
        For more details see Splink docs.

        In database this represents data as an integer counting number of days since
        1970-01-01 (Unix epoch).
        The input data can be either a string in YYYY-MM-DD format, or an
        integer of the number days since the epoch.

        Args:
            col_name (str): The name of the column to compare.
            input_is_string (bool): If True, the input dates are treated as strings
                and parsed to integers, and must be in ISO 8601 format.
            threshold (int): The maximum allowed difference between the two dates,
                in units specified by `date_metric`.
            metric (str): The unit of time to use when comparing the dates.
                Can be 'second', 'minute', 'hour', 'day', 'month', or 'year'.
        """
        super().__init__(
            col_name,
            input_is_string=input_is_string,
            threshold=threshold,
            metric=metric,
        )
        # need this to help mypy:
        self.col_expression: ColumnExpression

    @property
    def datetime_parsed_column_expression(self) -> CHColumnExpression:
        # convert existing ColumnExpression to our version,
        # and then apply parsing operation
        return CHColumnExpression.from_base_expression(
            self.col_expression
        ).parse_date_to_int()

    def create_sql(self, sql_dialect: SplinkDialect) -> str:
        self.col_expression.sql_dialect = sql_dialect
        if self.input_is_string:
            self.col_expression = self.datetime_parsed_column_expression

        col = self.col_expression

        # work in seconds as that's what parent uses, and we want to keep that machinery
        seconds_in_day = 86_400
        sql = (
            f"abs({col.name_l} - {col.name_r}) * {seconds_in_day} "
            f"<= {self.time_threshold_seconds}"
        )
        return sql


class PairwiseStringDistanceFunctionLevel(SplinkPairwiseStringDistanceFunctionLevel):
    def create_sql(self, sql_dialect: SplinkDialect) -> str:
        self.col_expression.sql_dialect = sql_dialect
        col = self.col_expression
        distance_function_name_transpiled = {
            "levenshtein": sql_dialect.levenshtein_function_name,
            "damerau_levenshtein": sql_dialect.damerau_levenshtein_function_name,
            "jaro_winkler": sql_dialect.jaro_winkler_function_name,
            "jaro": sql_dialect.jaro_function_name,
        }[self.distance_function_name]

        aggregator_func = {
            "min": sql_dialect.array_min_function_name,
            "max": sql_dialect.array_max_function_name,
        }[self._aggregator()]

        # order of the arguments is different in Clickhouse than tha expected by Splink
        # specifically the lambda must come first in Clickhouse
        # this is not fixable with UDF as having it in second argument in general
        # will cause Clickhouse parser to fail
        # also need to use a workaround to get 'flatten' equivalent for a single level
        return f"""{aggregator_func}(
                    {sql_dialect.array_transform_function_name}(
                        pair -> {distance_function_name_transpiled}(
                            pair[{sql_dialect.array_first_index}],
                            pair[{sql_dialect.array_first_index + 1}]
                        ),
                        arrayReduce(
                            'array_concat_agg',
                            {sql_dialect.array_transform_function_name}(
                                x -> {sql_dialect.array_transform_function_name}(
                                    y -> [x, y],
                                    {col.name_r}
                                ),
                                {col.name_l}
                            )
                        )
                    )
                ) {self._comparator()} {self.distance_threshold}"""



import logging

import pandas as pd
from clickhouse_connect.driver.client import Client
from splink.internals.database_api import DatabaseAPI

from ..custom_sql import days_since_epoch_sql
from ..dialect import ClickhouseDialect
from .dataframe import ClickhouseDataFrame

logger = logging.getLogger(__name__)


class ClickhouseAPI(DatabaseAPI[None]):
    sql_dialect = ClickhouseDialect()

    def __init__(
        self,
        client: Client,
        register_custom_udfs: bool = True,
    ):
        super().__init__()

        self.client = client
        self.set_union_default_mode()
        self._create_random_function()
        if register_custom_udfs:
            self._register_custom_udfs()

    def _table_registration(self, input, table_name) -> None:
        if isinstance(input, pd.DataFrame):
            sql = self._create_table_from_pandas_frame(input, table_name)
            self.client.query(sql)
            self.client.insert_df(table_name, input)
        elif isinstance(input, str):
            sql = (
                f"CREATE OR REPLACE TABLE {table_name} "
                "ORDER BY tuple() "
                f"AS SELECT * FROM {input}"
            )
            self.client.query(sql)
        else:
            raise TypeError(
                "ClickhouseAPI currently only accepts table names (str) "
                "or pandas DataFrames as inputs for table registration. "
                f"Received type {type(input)}"
            )

    def table_to_splink_dataframe(self, templated_name, physical_name):
        return ClickhouseDataFrame(templated_name, physical_name, self)

    def table_exists_in_database(self, table_name):
        sql = f"""
        SELECT table_name
        FROM information_schema.tables
        WHERE table_name = '{table_name}'
        AND table_schema = '{self.database}'
        """

        res = self.client.query(sql).result_set
        return len(res) > 0

    def _setup_for_execute_sql(self, sql: str, physical_name: str) -> str:
        self.delete_table_from_database(physical_name)
        sql = sql.replace("float8", "Float64")
        # workaround for https://github.com/ClickHouse/ClickHouse/issues/61004
        sql = sql.replace("count(*)", "count()")
        sql = sql.replace("COUNT(*)", "COUNT()")
        # TODO: very sorry for this
        # avoids 'double selection' issue in creating __splink__block_counts
        sql = sql.replace(", count_l, count_r,", ",")
        # some excessively brittle SQL replacements to hand Clickhouse name-resolution
        sql = sql.replace(
            "SELECT DISTINCT r.representative",
            "SELECT DISTINCT r.representative AS representative",
        )

        sql = f"CREATE TABLE {physical_name} ORDER BY tuple() AS {sql}"
        return sql

    def _execute_sql_against_backend(
        self, final_sql: str, templated_name: str = None, physical_name: str = None
    ):
        self.client.query(final_sql)

    @property
    def database(self) -> str:
        return self.client.database or "default"

    # Clickhouse can not handle a bare 'UNION' by default
    # we can set desired behaviour for the session by executing this
    def set_union_default_mode(self) -> None:
        self.client.command("SET union_default_mode = 'DISTINCT'")

    # alias random -> rand. Need this function for comparison viewer
    def _create_random_function(self) -> None:
        self.client.command("CREATE FUNCTION IF NOT EXISTS random AS () -> rand()")

    def _register_custom_udfs(self) -> None:
        self.client.command(
            f"""
            CREATE FUNCTION IF NOT EXISTS
                days_since_epoch AS
                (date_string) -> {days_since_epoch_sql}
            """
        )

    def _create_table_from_pandas_frame(self, df: pd.DataFrame, table_name: str) -> str:
        sql = f"CREATE OR REPLACE TABLE {table_name} ("

        first_col = True
        for column_name in df.columns:
            if not first_col:
                sql += ", "
            column = df[column_name]
            col_type = column.dtype
            first_col = False

            if pd.api.types.is_integer_dtype(col_type):
                sql += f"{column_name} Nullable(UInt32)"
            elif pd.api.types.is_float_dtype(col_type):
                sql += f"{column_name} Nullable(Float64)"
            elif pd.api.types.is_list_like(column[0]):
                sql += f"{column_name} Array(String)"
            elif pd.api.types.is_string_dtype(col_type):
                sql += f"{column_name} Nullable(String)"
            else:
                raise ValueError(f"Unknown data type {col_type}")

        sql += ") ENGINE MergeTree ORDER BY tuple()"

        return sql



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from splink.internals.input_column import InputColumn
from splink.internals.splink_dataframe import SplinkDataFrame

logger = logging.getLogger(__name__)
if TYPE_CHECKING:
    from .database_api import ClickhouseAPI


class ClickhouseDataFrame(SplinkDataFrame):
    db_api: ClickhouseAPI

    def __init__(self, df_name, physical_name, db_api):
        super().__init__(df_name, physical_name, db_api)

    @property
    def columns(self) -> list[InputColumn]:
        client = self.db_api.client

        sql = f"""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = '{self.physical_name}'
        AND table_schema = '{self.db_api.database}'
        """

        res = client.query(sql).named_results()
        cols = [r["column_name"] for r in res]

        return [InputColumn(c, sqlglot_dialect_str="clickhouse") for c in cols]

    def validate(self):
        if not self.db_api.table_exists_in_database(self.physical_name):
            raise ValueError(f"{self.physical_name} does not exist in the db provided.")

    def _drop_table_from_database(self, force_non_splink_table=False):
        self._check_drop_table_created_by_splink(force_non_splink_table)
        self.db_api.delete_table_from_database(self.physical_name)

    def as_record_dict(self, limit=None):
        sql = f"""
        SELECT *
        FROM {self.physical_name}
        """
        if limit:
            sql += f" LIMIT {limit}"
        sql += ";"

        res = self.db_api.client.query(sql)
        return list(res.named_results())



import logging

import chdb.dbapi as chdb_dbapi
import pandas as pd
from splink.internals.database_api import DatabaseAPI

from ..custom_sql import days_since_epoch_sql
from ..dialect import ClickhouseDialect
from .dataframe import ChDBDataFrame

logger = logging.getLogger(__name__)


class ChDBAPI(DatabaseAPI[None]):
    sql_dialect = ClickhouseDialect()

    def __init__(
        self,
        con: chdb_dbapi.Connection,
        schema: str = "splink",
        register_custom_udfs: bool = True,
    ):
        super().__init__()

        self.con = con
        self._db_schema = schema

        self._create_splink_schema()
        self._create_random_function()
        if register_custom_udfs:
            self._register_custom_udfs()

    def _table_registration(self, input, table_name):
        if isinstance(input, dict):
            input = pd.DataFrame(input)
        elif isinstance(input, list):
            input = pd.DataFrame.from_records(input)

        # chdb currently needs pandas indices to start at 0
        # see https://github.com/chdb-io/chdb/issues/282
        # reset the index if not the case, but otherwise leave alone
        # TODO: remove this workaround once chdb issue is resolved
        try:
            input[0]
        except KeyError:
            input = input.reset_index()
        cursor = self._get_cursor()
        sql = (
            f"CREATE OR REPLACE TABLE {self._db_schema}.{table_name} "
            "ORDER BY tuple() "
            f"AS SELECT * FROM Python(input);"
        )
        try:
            cursor.execute(sql)
        finally:
            # whatever happens, close the cursor
            self._reset_cursor(cursor)

    def table_to_splink_dataframe(self, templated_name, physical_name):
        return ChDBDataFrame(templated_name, physical_name, self)

    def table_exists_in_database(self, table_name):
        sql = f"""
        SELECT table_name
        FROM information_schema.tables
        WHERE table_name = '{table_name}';
        """

        cursor = self._get_cursor()
        try:
            cursor.execute(sql)
            table_name_row = cursor.fetchone()
        finally:
            self._reset_cursor(cursor)
        return table_name_row is not None

    def _setup_for_execute_sql(self, sql: str, physical_name: str) -> str:
        self.delete_table_from_database(physical_name)
        sql = sql.replace("float8", "Float64")
        # TODO: horrible hack
        # can't seem to set union_default_mode for some reason
        sql = sql.replace("UNION ALL", "__tmp__ua__")
        sql = sql.replace("UNION", "UNION DISTINCT")
        sql = sql.replace("__tmp__ua__", "UNION ALL")
        # workaround for https://github.com/ClickHouse/ClickHouse/issues/61004
        sql = sql.replace("count(*)", "count()")
        sql = sql.replace("COUNT(*)", "COUNT()")
        # TODO: very sorry for this
        # avoids 'double selection' issue in creating __splink__block_counts
        sql = sql.replace(", count_l, count_r,", ",")
        # some excessively brittle SQL replacements to hand Clickhouse name-resolution
        sql = sql.replace(
            "SELECT DISTINCT r.representative",
            "SELECT DISTINCT r.representative AS representative",
        )

        sql = f"CREATE TABLE {physical_name} ORDER BY tuple() AS {sql}"
        return sql

    def _execute_sql_against_backend(
        self, final_sql: str, templated_name: str = None, physical_name: str = None
    ) -> None:
        cursor = self._get_cursor()
        try:
            cursor.execute(final_sql)
        finally:
            self._reset_cursor(cursor)
        return None

    def _get_cursor(self) -> chdb_dbapi.cursors.DictCursor:
        return self.con.cursor(chdb_dbapi.cursors.DictCursor)

    def _reset_cursor(self, cursor):
        cursor.close()

    def _create_splink_schema(self):
        sql = f"""
        CREATE DATABASE IF NOT EXISTS {self._db_schema};
        USE {self._db_schema};
        """
        self._execute_sql_against_backend(sql)

    # alias random -> rand. Need this function for comparison viewer
    def _create_random_function(self) -> None:
        sql = "CREATE FUNCTION IF NOT EXISTS random AS () -> rand()"

        cursor = self._get_cursor()
        try:
            cursor.execute(sql)
        finally:
            self._reset_cursor(cursor)

    def _register_custom_udfs(self) -> None:
        sql = f"""
        CREATE FUNCTION IF NOT EXISTS
            days_since_epoch AS
            (date_string) -> {days_since_epoch_sql}
        """

        cursor = self._get_cursor()
        try:
            cursor.execute(sql)
        finally:
            self._reset_cursor(cursor)



from splink.internals.comparison_level_library import (
    AbsoluteTimeDifferenceLevel,
    ArrayIntersectLevel,
)
from splink.internals.dialects import SplinkDialect


class ClickhouseDialect(SplinkDialect):
    _dialect_name_for_factory = "clickhouse"

    @property
    def sql_dialect_str(self) -> str:
        return "clickhouse"

    @property
    def levenshtein_function_name(self) -> str:
        return "editDistanceUTF8"

    @property
    def damerau_levenshtein_function_name(self) -> str:
        return "damerauLevenshteinDistance"

    @property
    def jaro_winkler_function_name(self) -> str:
        return "jaroWinklerSimilarity"

    @property
    def jaro_function_name(self) -> str:
        return "jaroSimilarity"

    @property
    def jaccard_function_name(self) -> str:
        return "stringJaccardIndexUTF8"

    @property
    def array_first_index(self) -> int:
        return 1

    @property
    def array_min_function_name(self) -> str:
        return "arrayMin"

    @property
    def array_max_function_name(self) -> str:
        return "arrayMax"

    @property
    def array_transform_function_name(self) -> str:
        return "arrayMap"

    def _regex_extract_raw(
        self, name: str, pattern: str, capture_group: int = 0
    ) -> str:
        return f"regexpExtract({name}, '{pattern}', {capture_group})"

    def try_parse_date(self, name: str, date_format: str = None) -> str:
        if date_format is not None:
            raise ValueError(
                "Clickhouse dialect does not support custom `date_format`. "
                "To use a custom format you must manually construct "
                "the relevant expression."
            )
        return f"""parseDateTime64BestEffortOrNull({name})"""

    def try_parse_timestamp(self, name: str, timestamp_format: str = None) -> str:
        if timestamp_format is not None:
            raise ValueError(
                "Clickhouse dialect does not support custom `date_format`. "
                "To use a custom format you must manually construct "
                "the relevant expression."
            )
        return f"""parseDateTime64BestEffortOrNull({name})"""

    def absolute_time_difference(self, clc: AbsoluteTimeDifferenceLevel) -> str:
        # need custom solution as sqlglot keeps TIME_TO_UNIX which is not available
        clc.col_expression.sql_dialect = self
        col = clc.col_expression

        return (
            f"ABS(AGE('second', {col.name_l}, {col.name_r}))"
            f"<= {clc.time_threshold_seconds}"
        )

    def array_intersect(self, clc: ArrayIntersectLevel) -> str:
        clc.col_expression.sql_dialect = self
        col = clc.col_expression
        threshold = clc.min_intersection
        return f"""
        length(arrayIntersect({col.name_l}, {col.name_r})) >= {threshold}
        """.strip()

    def random_sample_sql(
        self, proportion, sample_size, seed=None, table=None, unique_id=None
    ):
        if proportion == 1.0:
            return ""
        if seed:
            raise NotImplementedError(
                "Clickhouse sampling does not support a random seed. "
                "Please remove the `seed` parameter."
            )

        sample_size = int(sample_size)

        # SAMPLE n only works when we have specified the method in table creation
        return f"WHERE randCanonical() < {proportion}"

    @property
    def infinity_expression(self):
        return "Inf"
